unit Controller.Device;

interface

uses
  System.SysUtils,
  System.Classes,
  System.Threading,
  mvcframework,
  mvcframework.Commons,
  LoggerProConfig,
  Model.BusinessObjects,
  Controller.Base;

type

  [MVCDoc('Resource that manages devices CRUD')]
  [MVCPath('/device')]
  TDeviceController = class(TBaseController)
  public
    [MVCDoc('Return the list of devices')]
    [MVCPath]
    [MVCHTTPMethod([httpGET])]
    procedure GetDevices;

    [MVCDoc('Return the device with the specified id')]
    [MVCPath('/($deviceid)')]
    [MVCHTTPMethod([httpGET])]
    function GetDeviceByID(deviceid: string; ARender: boolean = True): TDevice;

    [MVCDoc('Delete the device with the specified id')]
    [MVCPath('/($deviceid)')]
    [MVCHTTPMethod([httpDelete])]
    procedure DeleteDeviceByID(deviceid: string);

    [MVCDoc('Update/Create the device with the specified id')]
    [MVCPath('/($deviceid)')]
    [MVCHTTPMethod([httpPUT])]
    procedure UpdateDeviceByID(deviceid: string);

    [MVCDoc('Create a new device')]
    [MVCPath]
    [MVCHTTPMethod([httpPOST])]
    procedure CreateDevice(Context: TWebContext);

    [MVCPath('/authorize/($code)/($deviceid)')]
    [MVCHTTPMethod([httpPOST])]
    procedure AuthorizeDevice(code: string; deviceid: string);

  end;

implementation

{ TDevicesController }

uses
  mvcframework.Serializer.Intf, OneSignal.REST;

procedure TDeviceController.AuthorizeDevice(code: string; deviceid: string);
var
  LUrl: string;
  LDeviceFromDb: TDevice;
begin
  if GetDevicesService.Authorize(code, LUrl) then
  begin
    LDeviceFromDb := GetDeviceByID(deviceid, False);

    // Invia una notifica Push al dispositivo chiamando ONESIGNAL in un thread separato se nella tabella
    // il dispositivo ha gi associato un ExternalId che corrisponde al playerid di ONESIGNAL
    if Assigned(LDeviceFromDb) and (not LDeviceFromDb.ExternalId.IsEmpty) then
    begin
      Log.Debug(GetDataModule.Configuration.AppIdPushServer+' '+GetDataModule.Configuration.APIKeyPushServer+' '+LDeviceFromDb.ExternalId, 'SGSERVER');
      TOneSignal.SendPushToPlayer(GetDataModule.Configuration.AppIdPushServer, GetDataModule.Configuration.APIKeyPushServer, LDeviceFromDb.ExternalId);
    end;

    if Assigned(LDeviceFromDb) then LDeviceFromDb.Free;

    Render(200, LUrl)
  end
  else
    Render(401, 'Device Unauthorized');
end;

procedure TDeviceController.CreateDevice(Context: TWebContext);
var
  Device: TDevice;
begin
  Device := Context.Request.BodyAs<TDevice>;
  try
    GetDevicesService.Add(Device);
    Render(201, 'Device Created');
  finally
    Device.Free;
  end;
end;

procedure TDeviceController.DeleteDeviceByID(deviceid: string);
var
  Device: TDevice;
begin
  GetDevicesService.StartTransaction;
  try
    Device := GetDevicesService.GetByID(deviceid);
    try
      GetDevicesService.Delete(Device);
    finally
      Device.Free;
    end;
    GetDevicesService.Commit;
    Render(200, 'Device Deleted');
  except
    GetDevicesService.Rollback;
    raise;
  end;
end;

procedure TDeviceController.GetDevices;
begin
  Render<TDevice>(GetDevicesService.GetAll);
end;

procedure TDeviceController.UpdateDeviceByID(deviceid: string);
var
  LDeviceFromBody: TDevice;
  LDeviceFromDb: TDevice;
begin
  LDeviceFromBody := Context.Request.BodyAs<TDevice>;
  LDeviceFromDb := GetDeviceByID(deviceid, False);
  if Assigned(LDeviceFromDb) then
  begin
    LDeviceFromDb.DeviceToken := LDeviceFromBody.DeviceToken;
    LDeviceFromDb.Description := LDeviceFromBody.Description;
    LDeviceFromDb.DeviceType := LDeviceFromBody.DeviceType;
    LDeviceFromDb.DeviceOs := LDeviceFromBody.DeviceOs;
    LDeviceFromDb.DeviceOsVersion := LDeviceFromBody.DeviceOsVersion;
    LDeviceFromDb.DeviceModel := LDeviceFromBody.DeviceModel;

    GetDevicesService.Update(LDeviceFromDb);

    LDeviceFromDb.ExternalId := TOneSignal.UpdatePlayer(GetDataModule.Configuration.AppIdPushServer, GetDataModule.Configuration.APIKeyPushServer, LDeviceFromDb.DeviceToken, LDeviceFromDb.ExternalId, LDeviceFromDb.DeviceType, LDeviceFromDb.DeviceModel, LDeviceFromDb.DeviceOs+' '+LDeviceFromDb.DeviceOsVersion);
    GetDevicesService.UpdateExternalId(LDeviceFromDb);
    LDeviceFromDb.Free;

    Render(200, 'Device Updated');

  end
  else
  begin
    LDeviceFromBody.DeviceId := deviceid;

    GetDevicesService.Add(LDeviceFromBody);

    LDeviceFromBody.ExternalId := TOneSignal.UpdatePlayer(GetDataModule.Configuration.AppIdPushServer, GetDataModule.Configuration.APIKeyPushServer, LDeviceFromBody.DeviceToken, '', LDeviceFromBody.DeviceType, LDeviceFromBody.DeviceModel, LDeviceFromBody.DeviceOs+' '+LDeviceFromBody.DeviceOsVersion);
    GetDevicesService.UpdateExternalId(LDeviceFromBody);
    LDeviceFromBody.Free;

    Render(200, 'Device Created');
  end;
end;

function TDeviceController.GetDeviceByID(deviceid: string; ARender: boolean = True): TDevice;
begin
  try
    Result := GetDevicesService.GetByID(deviceid, False);
    if ARender then
      Render(Result);
  except
    on E: Exception do
    begin
      raise EMVCException.Create(E.Message, '', 0, 404);
    end
    else
      raise;
  end;
end;

end.
