unit Service.Services;

interface

uses
  System.Generics.Collections,
  System.SysUtils,
  Model.BusinessObjects,
  Server.DataModule,
  Global.Utility;

type

  TServiceBase = class abstract
  strict protected
    FDM: TDM;
  public
    constructor Create(AdmMain: TDM); virtual;
    procedure Commit;
    procedure Rollback;
    procedure StartTransaction;
  end;

  TDevicesService = class(TServiceBase)
  public
    function GetAll: TObjectList<TDevice>;
    function GetByID(const ADeviceID: string; const ARaiseErr: boolean = True): TDevice;
    procedure Delete(ADevice: TDevice);
    procedure Add(ADevice: TDevice);
    procedure Update(ADevice: TDevice);
    procedure UpdateExternalId(ADevice: TDevice);
    function Authorize(ACode: string; out oUrl: string): boolean;
  end;

implementation

uses
  FireDAC.Stan.Option, FireDAC.Comp.Client, FireDAC.Stan.Param,
  MVCFramework.FireDAC.Utils, MVCFramework.DataSet.Utils,
  MVCFramework.Serializer.Commons,
  Server.Configuration;

{ TArticoliService }

procedure TDevicesService.Add(ADevice: TDevice);
var
  Cmd: TFDCustomCommand;
  LTq: TFDQuery;
begin
  ADevice.CheckInsert;
  Cmd := FDM.updDevices.Commands[arInsert];
  TFireDACUtils.ObjectToParameters(Cmd.Params, ADevice, 'NEW_');
  Cmd.Execute;

  LTq := FDM.NewQuery;
  try
    LTq.Sql.Clear;
    Ltq.Sql.Add('SELECT LAST_INSERT_ID() AS LASTID');
    LTq.Open;
    ADevice.ID := LTq.FieldByName('LASTID').AsInteger;
    LTq.Close;
  finally
    LTq.Free;
  end;
end;

function TDevicesService.Authorize(ACode: string; out oUrl: string): boolean;
begin
  Result := False;
  oUrl := '';

  if LowerCase(ACode)='delphiday' then
  begin
    Result := True;
    oUrl := 'www.dnasoftware.it';
  end;
end;

procedure TDevicesService.Delete(ADevice: TDevice);
var
  Cmd: TFDCustomCommand;
begin
  ADevice.CheckDelete;
  Cmd := FDM.updDevices.Commands[arDelete];
  TFireDACUtils.ObjectToParameters(Cmd.Params, ADevice, 'OLD_');
  Cmd.Execute;
end;

function TDevicesService.GetAll: TObjectList<TDevice>;
begin
  FDM.dsDevices.Open('SELECT * FROM DEVICES ORDER BY ID');
  Result := FDM.dsDevices.AsObjectList<TDevice>;
  FDM.dsDevices.Close;
end;

function TDevicesService.GetByID(const ADeviceID: string; const ARaiseErr: boolean): TDevice;
var
  LTq: TFDQuery;
begin
  Result := nil;

  LTq := FDM.NewQuery;
  try
    LTq.Sql.Clear;
    Ltq.Sql.Add('SELECT * FROM DEVICES WHERE DEVICEID = '+ADeviceId.QuotedString);
    LTq.Open;
    if not LTq.Eof then
      Result := LTq.AsObject<TDevice>();

    if (not Assigned(Result)) and ARaiseErr then
      raise Exception.Create('Device not found');

    LTq.Close;
  finally
    LTq.Free;
  end;
end;

procedure TDevicesService.Update(ADevice: TDevice);
var
  Cmd: TFDCustomCommand;
begin
  ADevice.CheckUpdate;
  Cmd := FDM.updDevices.Commands[arUpdate];
  TFireDACUtils.ObjectToParameters(Cmd.Params, ADevice, 'NEW_');
  Cmd.ParamByName('OLD_DEVICEID').AsString := ADevice.DeviceId;
  Cmd.Execute;
  if Cmd.RowsAffected <> 1 then
    raise Exception.Create('Device not found');
end;

procedure TDevicesService.UpdateExternalId(ADevice: TDevice);
var
  LTq: TFDQuery;
begin
  LTq := FDM.NewQuery;
  try
    LTq.Sql.Clear;
    LTq.Sql.Add('UPDATE DEVICES SET EXTERNALID='+ADevice.ExternalId.QuotedString+' WHERE DEVICEID='+ADevice.DeviceId.QuotedString);
    LTq.ExecSql;
  finally
    LTq.Free;
  end;
end;

{ TServiceBase }

procedure TServiceBase.Commit;
begin
  FDM.CommitTransaction;
end;

constructor TServiceBase.Create(AdmMain: TDM);
begin
  inherited Create;
  FDM := AdmMain;
end;

procedure TServiceBase.Rollback;
begin
  FDM.RollbackTransaction;
end;

procedure TServiceBase.StartTransaction;
begin
  FDM.BeginTransaction;
end;

end.

